package io.tswf.minecraft.blockventure.core.api.registry.dynamic

import com.google.auto.service.AutoService
import com.mojang.serialization.Lifecycle
import groovy.transform.CompileStatic
import groovy.transform.Memoized
import net.minecraft.registry.*
import net.minecraft.structure.processor.StructureProcessorLists
import net.minecraft.world.gen.carver.ConfiguredCarvers
import net.minecraft.world.gen.feature.ConfiguredFeatures
import net.minecraft.world.gen.feature.PlacedFeatures

import java.util.concurrent.ConcurrentHashMap

@CompileStatic
class StaticDynamicRegistryLookup {

    static Map<RegistryKey, RegistryBuilder.BootstrapFunction> bootstrapRegistries = new ConcurrentHashMap<>()

    @Lazy
    static List<ILookupSetupListener> setupListeners = ServiceLoader.load(ILookupSetupListener)
            .toList()

    @Memoized
    static DynamicRegistryManager getStaticRegistryManager() {
        return DynamicRegistryManager.of(Registries.REGISTRIES)
    }

    /**
     * Собирает статические Registry и подпихивает по-возможности весь контент из датапаков, до какого можно достучаться.
     * @return
     */
    @Memoized
    static RegistryWrapper.WrapperLookup getEnrichedRegistryLookup() {
        def allStaticRegistries = Registries.REGISTRIES.stream()
                .toList()

        def allStaticRegistryKeys = allStaticRegistries.map { it.key }

        def allDynamicRegistryKeys = RegistryKeys.fields.stream()
                .filter { it.modifierInfo.with { isPublic() && isStatic() && isFinal() } }
                .filter { it.type == RegistryKey }
                .map { (RegistryKey) it.get(null) }
                .filter { it !in allStaticRegistryKeys }
                .distinct()
                .toList()

        def allRegistries = new ArrayList<Registry>()
        allRegistries.addAll()

        for (def dynamicRegistryKey : allDynamicRegistryKeys) {
            allRegistries << new SimpleRegistry<>(dynamicRegistryKey, Lifecycle.stable())
        }

        allRegistries.removeIf { bootstrapRegistries.containsKey(it.key) }

        def dynRegistries = new DynamicRegistryManager.ImmutableImpl((List) allRegistries)
        def regBuilder = new RegistryBuilder()

        for (def bootstrapEntry : bootstrapRegistries) {
            regBuilder.addRegistry(bootstrapEntry.key, bootstrapEntry.value)
        }

        return regBuilder.createWrapperLookup(dynRegistries)
    }

    static <T> void registerBootstrap(RegistryKey<? extends Registry<T>> registryRef, RegistryBuilder.BootstrapFunction<T> bootstrapFunction) {
        bootstrapRegistries[registryRef] = bootstrapFunction
    }

    static {
        for (def setupListener : setupListeners) {
            setupListener.handleLookupSetup()
        }
    }

    static interface ILookupSetupListener {
        void handleLookupSetup()
    }

    @AutoService(ILookupSetupListener)
    static class WorldGenLookupSetupListener implements ILookupSetupListener {

        @Override
        void handleLookupSetup() {
            registerBootstrap(RegistryKeys.PLACED_FEATURE, { PlacedFeatures.bootstrap(it) })
            registerBootstrap(RegistryKeys.CONFIGURED_CARVER, { ConfiguredCarvers.bootstrap(it) })
            registerBootstrap(RegistryKeys.CONFIGURED_FEATURE, { ConfiguredFeatures.bootstrap(it) })
            registerBootstrap(RegistryKeys.PROCESSOR_LIST, { StructureProcessorLists.bootstrap(it) })
        }
    }
}